#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rating.h"


/*-------------------------------------------------------------------*/
/* 変数                                                              */
/*-------------------------------------------------------------------*/

/* γ */
float tactical_features[TACTICAL_FEATURE];
float border_distance[BORDER_DISTANCE_MAX];
float move_distance_1[MOVE_DISTANCE_MAX];
float move_distance_2[MOVE_DISTANCE_MAX];
float pattern3[PATTERN3_MAX];
float diamond[DIAMOND_MAX];

/* 盤端との距離のレート */
float border_rate[BOARD_MAX];

/* 初期値 */
struct rate_set init_rate_set[BOARD_MAX];


/*-------------------------------------------------------------------*/
/* 初期値                                                            */
/*-------------------------------------------------------------------*/
void
InitializeRating( void )
{
	int x, y;
	int pos;

	/* γ読み込み */
	InputGamma();

	/* 盤端との距離のレート*/
	for (y=BOARD_START; y<=BOARD_END; y++) {
		for (x=BOARD_START; x<=BOARD_END; x++) {
			pos = POS(x,y);
			border_rate[pos] = 
				border_distance[border_dis_x[pos]] * border_distance[border_dis_y[pos]];
			init_rate_set[pos].list[0] = 1.0;
			init_rate_set[pos].list[1] = 1.0;
		}
	}
}


/*-------------------------------------------------------------------*/
/* 初期設定                                                          */
/*-------------------------------------------------------------------*/
void
ClearRate( struct game *game )
{
	int *rate = game->rate;
	int *sum_rate_column = game->sum_rate_column;
	int *sum_rate = &game->sum_rate;
	int x, y, pos;
	double gamma;

	/* 初期化 */
	memset(rate, 0, sizeof(int)*BOARD_MAX);
	memset(sum_rate_column, 0, sizeof(int)*BOARD_SIZE);
	*sum_rate = 0;

	for( y=BOARD_START; y<=BOARD_END; y++ ){
		for( x=BOARD_START; x<=BOARD_END; x++ ){
			pos = POS(x, y);
			gamma = diamond[game->dia[pos]];
			if(gamma == 1.0) gamma = pattern3[game->pat3[pos]];
			gamma *= border_rate[pos];
			rate[pos] = (int)(gamma)+1;
			*sum_rate += rate[pos];
			sum_rate_column[board_y[pos]] += rate[pos];
		}
	}
}


/*-------------------------------------------------------------------*/
/* 着手( rating )                                                    */
/*-------------------------------------------------------------------*/
int
RatingMove( struct game *game, int color, struct mt19937p *mt )
{
	int *rate = game->rate;
	int *sum_rate_column = game->sum_rate_column;
	int *sum_rate = &game->sum_rate;
	int y, pos;
	int rand_num;

	/* 初期化 */
	memset(rate, 0, sizeof(int)*BOARD_MAX);
	memset(sum_rate_column, 0, sizeof(int)*BOARD_SIZE);
	*sum_rate = 0;

	/* レーティング */
	Rating( game, color );

	if (*sum_rate == 0) return PASS;

	rand_num = (genrand(mt) % (*sum_rate))+1;

	y=BOARD_START;
	while(rand_num > sum_rate_column[y]){
		rand_num -= sum_rate_column[y++];
	}

	pos = POS(BOARD_START,y);
	do{
		rand_num-=rate[pos];
		if(rand_num <= 0) break;
		pos++;
	}while( 1 );

	return pos;
}


/*-------------------------------------------------------------------*/
/* Rating                                                            */
/*-------------------------------------------------------------------*/
void
Rating( struct game *game, int color )
{
	int i, pos;
	struct rate_set rate_set[BOARD_MAX];

	/* 初期化 */
	memcpy(rate_set, init_rate_set, sizeof(struct rate_set)*BOARD_MAX);

	/* 特徴の判定 */
	CheckFeatures( game, color, rate_set );

	if (game->ko_move == game->moves-2) {
		CheckCaptureAfterKo( game, color, rate_set );
	}

	for ( i=0; i<PURE_BOARD_MAX; i++ ) {
		pos = onboard_pos[i];
		if (game->candidates[pos] && CheckLegalAndEye( game, color, pos ) == TRUE) {
			CheckSelfAtari( game, color, pos, rate_set );
			if(rate_set[pos].list[0] == 1.0) CheckCapture( game, color, pos, rate_set );
			if(rate_set[pos].list[1] == 1.0) CheckAtari( game, color, pos, rate_set );
			UpdateRate( game, pos, rate_set );
		}
	}
}


/*-------------------------------------------------------------------*/
/* レートの更新                                                      */
/*-------------------------------------------------------------------*/
void
UpdateRate( struct game *game, int pos, struct rate_set *rate_set )
{
	int pm1 = PASS, pm2 = PASS;
	float gamma;

	pm1 = game->record[game->moves-1].pos;
	if (game->moves > 2) pm2 = game->record[game->moves-2].pos;

	gamma = diamond[game->dia[pos]];
	if(gamma == 1.0) gamma = pattern3[game->pat3[pos]];
	gamma *= rate_set[pos].list[0];
	gamma *= rate_set[pos].list[1];
	if (pm1 != PASS) gamma *= move_distance_1[DIS(pm1, pos)];
	if (pm2 != PASS) gamma *= move_distance_2[DIS(pm2, pos)];
	gamma *= border_rate[pos];

	game->rate[pos] = (int)(gamma)+1;
	game->sum_rate += game->rate[pos];
	game->sum_rate_column[board_y[pos]] += game->rate[pos];
}


/*-------------------------------------------------------------------*/
/* 特徴の判定                                                        */
/*-------------------------------------------------------------------*/
void
CheckFeatures( struct game *game, int color, struct rate_set *rate_set )
{
	struct string *string = game->string;
	struct string_liberties *string_liberties = game->string_liberties;
	struct string_neighbors *string_neighbors = game->string_neighbors;
	int previous_move = game->record[game->moves-1].pos;
	int i;
	int str;

	if(previous_move == PASS) return;

	/* 連識別番号の更新 */
	game->string_mark++;

	/* 上 */
	if (game->board[NORTH(previous_move)].color == color) {
		str = game->board[NORTH(previous_move)].string_number;
		if (string[str].liberties == 1) {
			string[str].mark = game->string_mark;
			/* save new atari by extending */
			rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_SAVE_EXTENSION];
			/* save new atari by capturing */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 1) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[0] = tactical_features[F_SAVE_CAPTURE];
				}
			}
		} else if (string[str].liberties == 2) {
			string[str].mark = game->string_mark+1;
			/* 2point semeai */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 2) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[1] = tactical_features[F_2POINT_SEMEAI];
					rate_set[string_liberties[string_neighbors[str].list[i]].list[1]].list[1] = tactical_features[F_2POINT_SEMEAI];
				}
			}
		}
	}

	/* 右 */
	if (game->board[EAST(previous_move)].color == color) {
		str = game->board[EAST(previous_move)].string_number;
		if (string[str].liberties == 1 && string[str].mark != game->string_mark) {
			string[str].mark = game->string_mark;
			/* save new atari by extending */
			rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_SAVE_EXTENSION];
			/* save new atari by capturing */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 1) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[0] = tactical_features[F_SAVE_CAPTURE];
				}
			}
		} else if (string[str].liberties == 2 && string[str].mark != game->string_mark+1) {
			string[str].mark = game->string_mark+1;
			/* 2point semeai */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 2) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[1] = tactical_features[F_2POINT_SEMEAI];
					rate_set[string_liberties[string_neighbors[str].list[i]].list[1]].list[1] = tactical_features[F_2POINT_SEMEAI];
				}
			}
		}
	}

	/* 下 */
	if (game->board[SOUTH(previous_move)].color == color) {
		str = game->board[SOUTH(previous_move)].string_number;
		if (string[str].liberties == 1 && string[str].mark != game->string_mark) {
			string[str].mark = game->string_mark;
			/* save new atari by extending */
			rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_SAVE_EXTENSION];
			/* save new atari by capturing */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 1) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[0] = tactical_features[F_SAVE_CAPTURE];
				}
			}
		} else if (string[str].liberties == 2 && string[str].mark != game->string_mark+1) {
			string[str].mark = game->string_mark+1;
			/* 2point semeai */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 2) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[1] = tactical_features[F_2POINT_SEMEAI];
					rate_set[string_liberties[string_neighbors[str].list[i]].list[1]].list[1] = tactical_features[F_2POINT_SEMEAI];
				}
			}
		}
	}

	/* 左 */
	if (game->board[WEST(previous_move)].color == color) {
		str = game->board[WEST(previous_move)].string_number;
		if (string[str].liberties == 1 && string[str].mark != game->string_mark) {
			/* save new atari by extending */
			rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_SAVE_EXTENSION];
			/* save new atari by capturing */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 1) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[0] = tactical_features[F_SAVE_CAPTURE];
				}
			}
		} else if (string[str].liberties == 2 && string[str].mark != game->string_mark+1) {
			/* 2point semeai */
			for (i=0; i<string[str].neighbors; i++) {
				if (string[string_neighbors[str].list[i]].liberties == 2) {
					rate_set[string_liberties[string_neighbors[str].list[i]].list[0]].list[1] = tactical_features[F_2POINT_SEMEAI];
					rate_set[string_liberties[string_neighbors[str].list[i]].list[1]].list[1] = tactical_features[F_2POINT_SEMEAI];
				}
			}
		}
	}

	/* 連識別番号の更新 */
	game->string_mark++;
}


/*-------------------------------------------------------------------*/
/* Capture after ko                                                  */
/*-------------------------------------------------------------------*/
void
CheckCaptureAfterKo( struct game *game, int color, struct rate_set *rate_set )
{
	struct string *string = game->string;
	struct string_liberties *string_liberties = game->string_liberties;
	int other = OTHER_COLOR(color);
	int previous_move_2 = game->record[game->moves-2].pos;
	int str;

	/* 連識別番号の更新 */
	game->string_mark++;

	/* 上 */
	if (game->board[NORTH(previous_move_2)].color == other) {
		str = game->board[NORTH(previous_move_2)].string_number;
		if (string[str].liberties == 1) {
			string[str].mark = game->string_mark;
			if(rate_set[string_liberties[str].list[0]].list[0] == 0){
				rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_CAPTURE_AFTER_KO];
			}
		}
	}

	/* 右 */
	if (game->board[EAST(previous_move_2)].color == other) {
		str = game->board[EAST(previous_move_2)].string_number;
		if (string[str].liberties == 1 && string[str].mark != game->string_mark) {
			string[str].mark = game->string_mark;
			if(rate_set[string_liberties[str].list[0]].list[0] == 0){
				rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_CAPTURE_AFTER_KO];
			}
		}
	}

	/* 下 */
	if (game->board[SOUTH(previous_move_2)].color == other) {
		str = game->board[SOUTH(previous_move_2)].string_number;
		if (string[str].liberties == 1 && string[str].mark != game->string_mark) {
			string[str].mark = game->string_mark;
			if(rate_set[string_liberties[str].list[0]].list[0] == 0){
				rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_CAPTURE_AFTER_KO];
			}
		}
	}

	/* 左 */
	if (game->board[WEST(previous_move_2)].color == other) {
		str = game->board[WEST(previous_move_2)].string_number;
		if (string[str].liberties == 1 && string[str].mark != game->string_mark) {
			if(rate_set[string_liberties[str].list[0]].list[0] == 0){
				rate_set[string_liberties[str].list[0]].list[0] = tactical_features[F_CAPTURE_AFTER_KO];
			}
		}
	}
}


/*-------------------------------------------------------------------*/
/* 自己アタリ                                                        */
/*-------------------------------------------------------------------*/
void
CheckSelfAtari( struct game *game, int color, int pos, struct rate_set *rate_set )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	int other = OTHER_COLOR(color);
	int libs = nb4_empty[game->pat3[pos]];

	/* 空点 */
	if (libs >= 2) return;

	/* 上を調べる */
	if (board[NORTH(pos)].color == color) {
		if (string[board[NORTH(pos)].string_number].liberties > 1) {
			libs += string[board[NORTH(pos)].string_number].liberties-1;
			if (libs >= 2) return;
		}
	} else if (board[NORTH(pos)].color == other
		&& string[board[NORTH(pos)].string_number].liberties == 1) {
			return;
	}

	/* 右を調べる */
	if (board[EAST(pos)].color == color) {
		if (string[board[EAST(pos)].string_number].liberties > 1) {
			libs += string[board[EAST(pos)].string_number].liberties-1;
			if (libs >= 2) return;
		}
	} else if (board[EAST(pos)].color == other
		&& string[board[EAST(pos)].string_number].liberties == 1) {
			return;
	}

	/* 下を調べる */
	if (board[SOUTH(pos)].color == color) {
		if (string[board[SOUTH(pos)].string_number].liberties > 1) {
			libs += string[board[SOUTH(pos)].string_number].liberties-1;
			if (libs >= 2) return;
		}
	} else if (board[SOUTH(pos)].color == other
		&& string[board[SOUTH(pos)].string_number].liberties == 1) {
			return;
	}

	/* 左を調べる */
	if (board[WEST(pos)].color == color) {
		if (string[board[WEST(pos)].string_number].liberties > 1) {
			libs += string[board[WEST(pos)].string_number].liberties-1;
			if (libs >= 2) return;
		}
	} else if (board[WEST(pos)].color == other
		&& string[board[WEST(pos)].string_number].liberties == 1) {
			return;
	}

	/* 自己アタリ */
	rate_set[pos].list[0] = tactical_features[F_SELF_ATARI];
}


/*-------------------------------------------------------------------*/
/* トリの判定                                                        */
/*-------------------------------------------------------------------*/
void
CheckCapture( struct game *game, int color, int pos, struct rate_set *rate_set )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	int other = OTHER_COLOR(color);

	/* 上を調べる */
	if (board[NORTH(pos)].color == other) {
		if (string[board[NORTH(pos)].string_number].liberties == 1) {
			rate_set[pos].list[0] = tactical_features[F_CAPTURE];
			return;
		}
	}

	/* 右を調べる */
	if (board[EAST(pos)].color == other) {
		if (string[board[EAST(pos)].string_number].liberties == 1) {
			rate_set[pos].list[0] = tactical_features[F_CAPTURE];
			return;
		}
	}

	/* 下を調べる */
	if (board[SOUTH(pos)].color == other) {
		if (string[board[SOUTH(pos)].string_number].liberties == 1) {
			rate_set[pos].list[0] = tactical_features[F_CAPTURE];
			return;
		}
	}

	/* 左を調べる */
	if (board[WEST(pos)].color == other) {
		if (string[board[WEST(pos)].string_number].liberties == 1) {
			rate_set[pos].list[0] = tactical_features[F_CAPTURE];
			return;
		}
	}
}


/*-------------------------------------------------------------------*/
/* アタリの判定                                                      */
/*-------------------------------------------------------------------*/
void
CheckAtari( struct game *game, int color, int pos, struct rate_set *rate_set )
{
	struct stone *board = game->board;
	struct string *string = game->string;
	int other = OTHER_COLOR(color);

	/* 上を調べる */
	if (board[NORTH(pos)].color == other) {
		if (string[board[NORTH(pos)].string_number].liberties == 2) {
			rate_set[pos].list[1] = tactical_features[F_ATARI];
			return;
		}
	}

	/* 右を調べる */
	if (board[EAST(pos)].color == other) {
		if (string[board[EAST(pos)].string_number].liberties == 2) {
			rate_set[pos].list[1] = tactical_features[F_ATARI];
			return;
		}
	}

	/* 下を調べる */
	if (board[SOUTH(pos)].color == other) {
		if (string[board[SOUTH(pos)].string_number].liberties == 2) {
			rate_set[pos].list[1] = tactical_features[F_ATARI];
			return;
		}
	}

	/* 左を調べる */
	if (board[WEST(pos)].color == other) {
		if (string[board[WEST(pos)].string_number].liberties == 2) {
			rate_set[pos].list[1] = tactical_features[F_ATARI];
			return;
		}
	}
}


/*------------------------------------------------------------------*/
/* γ読み込み                                                       */
/*------------------------------------------------------------------*/
void
InputGamma( void )
{
	/* 戦術的特徴 */
	InputTxtFLT("../params/TacticalFeature.txt", tactical_features, TACTICAL_FEATURE);

	/* 盤端との距離 */
	InputTxtFLT("../params/BorderDistance.txt", border_distance, BORDER_DISTANCE_MAX);

	/* 直前の手との距離 */
	InputMoveDistance("../params/MoveDistance1.txt", move_distance_1);

	/* 二手前の手との距離 */
	InputMoveDistance("../params/MoveDistance2.txt", move_distance_2);

	/* 3x3パターン */
	InputTxtFLT("../params/Pattern3.txt", pattern3, PATTERN3_MAX);

	/* ひし形パターン */
	InputDiamond("../params/Diamond.txt");
}


/*------------------------------------------------------------------*/
/* テキスト入力 (float)                                             */
/*------------------------------------------------------------------*/
void
InputTxtFLT( char *filename, float *ap, int array_size ) {
	int i;
	char s[256];
	FILE *fp;

	if(fopen_s(&fp, filename, "r") != 0){
		fprintf(stderr, "can not open -%s-\n", filename);
		exit(EXIT_FAILURE);
	}
	for(i=0; i<array_size; i++){
		fgets( s, 255, fp);
		ap[i] = (float)atof(s);
	}
	fclose( fp );
}


/*------------------------------------------------------------------*/
/* MoveDistance γ読み込み                                          */
/*------------------------------------------------------------------*/
void
InputMoveDistance( char *filename, float *ap )
{
	int i;
	int max_dis = 17;
	char s[256];
	FILE *fp;

	if(fopen_s(&fp, filename, "r") != 0){
		fprintf(stderr, "can not open -%s-\n", filename);
		exit(EXIT_FAILURE);
	}
	for(i=0; i<=max_dis; i++){
		fgets( s, 255, fp);
		ap[i] = (float)atof(s);
	}
	fclose( fp );
	for(; i<MOVE_DISTANCE_MAX; i++){
		ap[i] = ap[max_dis];
	}
}


/*------------------------------------------------------------------*/
/* Diamond Pattern γ読み込み                                       */
/*------------------------------------------------------------------*/
void
InputDiamond( char *filename )
{
	int i;
	FILE *fp;
	int index;
	float rate;

	for(i=0; i<DIAMOND_MAX; i++) diamond[i] = 1.0;
	if(fopen_s(&fp, filename, "r") != 0){
		fprintf(stderr, "can not open -%s-\n", filename);
		return;
	}
	while(fscanf_s( fp , "%d%f" , &index , &rate ) != EOF ) {
		diamond[index] = rate;
	}
	fclose( fp );
}

